
/*
 * Required headers
 */
#include <sys/types.h> /* open */
#include <sys/stat.h> /* open */
#include <fcntl.h> /* open */
#include <unistd.h> /* write, close, unlink */
#include <stdlib.h> /* malloc, realloc, free */

#include "headers.h" /* system-specific stuff */
#include "messages.h" /* ERROR, INFO */

/*
 * Prototypes
 */
#include "fichiers.h"

/*
 * Functions
 */
#ifdef TEST_SUITE

#define TEST_FILE "/tmp/transaction"
#define TEST_STRING "foo bar baz quuuuux 1234567890 WOOGA ZoY"

int main( void )
{
    transaction_t *p_trans;

    /* Write test */
    if( EcrireTransaction( TEST_FILE, TEST_STRING, strlen( TEST_STRING ) ) )
    {
        return -1;
    }

    INFO( "Wrote transaction `%s' to `%s'", TEST_STRING, TEST_FILE );

    /* Read test */
    p_trans = LireTransaction( TEST_FILE );
    if( p_trans == NULL )
    {
        return -1;
    }

    INFO( "Read transaction `%s' from `%s'", p_trans->p_data, TEST_FILE );

    /* Clean up */
    if( unlink( TEST_FILE ) )
    {
        ERROR( "Couldn't unlink `%s'", TEST_FILE );
        return -1;
    }

    /* Everything was OK */
    INFO( "Test successful" );
    return 0;
}
#endif

int EcrireTransaction( char *psz_file, unsigned char *p_data, ssize_t i_size )
{
    /* Open the file in write mode, and quit on error */
    int i_fd = open( psz_file, O_CREAT | O_WRONLY | O_APPEND, 0777 );

    if( i_fd < 0 )
    {
        ERROR( "EcrireTransaction: couldn't open %s", psz_file );
        return i_fd;
    }

    /* Try to write the header, otherwise return with an error */
    if( write( i_fd, "{:", 2 ) != 2 )
    {
        ERROR( "EcrireTransaction: failed writing header" );
        close( i_fd );
        return -1;
    }

    /* Try to write the data, otherwise return with an error */
    if( write( i_fd, p_data, i_size ) != i_size )
    {
        ERROR( "EcrireTransaction: failed writing transaction data" );
        close( i_fd );
        return -1;
    }

    /* Try to write the trailer, otherwise return with an error */
    if( write( i_fd, ":}\n", 3 ) != 3 )
    {
        ERROR( "EcrireTransaction: failed writing trailer" );
        close( i_fd );
        return -1;
    }

    /* Close the file and terminate all write commands */
    if( close( i_fd ) != 0 )
    {
        ERROR( "EcrireTransaction: error when closing %s", psz_file );
        return -1;
    }

    /* Everything worked properly, return */
    return 0;
}

transaction_t * LireTransaction( char *psz_file )
{
    /* Buffer data */
    char p_buf[ READ_ONCE ];
    int  i_buf_len = 0;
    int  i_buf_pos = 0;
    char i_dummy;
    char i_colon = ':';

    /* Transaction data */
    transaction_t *p_trans = MallocTransaction( 0 );

    /* Open the file in read mode, and quit on error */
    int i_fd = open( psz_file, O_RDONLY );
    if( i_fd < 0 )
    {
        ERROR( "LireTransaction: failed opening `%s' for reading", psz_file );
        return NULL;
    }

    /* Finite state automata for file parsing, just for fun */

begin: /* wait loop, stay here until we find '{' */
    READCHAR( &i_dummy );
    if( i_dummy == '{' ) goto state2;
    goto begin;

state2: /* '{' was found, we enter the main loop if the next one is ':' */
    READCHAR( &i_dummy );
    if( i_dummy == ':' ) goto state3;
    goto begin;

state3: /* main loop, copy read char only if not ':' */
    READCHAR( &i_dummy );
    if( i_dummy == ':' ) goto state4;
    ADDCHAR( i_dummy );
    goto state3;

state4: /* ':' was found, we end only if the next one is '}' */
    READCHAR( &i_dummy );
    if( i_dummy == '}' ) goto end;
    /* False positive, store the previous ':' and the current char */
    ADDCHAR( i_colon );
    ADDCHAR( i_dummy );
    goto state3;

error:
    FreeTransaction( p_trans );
    p_trans = NULL;
    goto end;

end:

    /* End of finite state automata */

    /* Return what we read, no need to check for close errors */
    close( i_fd );
    return p_trans;
}

/* Transaction allocation */
transaction_t * MallocTransaction( ssize_t i_size )
{
    transaction_t *p_trans;
    ssize_t i_allocsize = ( i_size + DATA_ALIGN - 1 ) & ~( DATA_ALIGN - 1 );

    p_trans = malloc( sizeof( transaction_t )
                       + i_allocsize * sizeof( unsigned char ) );

    p_trans->i_size = i_size;
    p_trans->p_data = ( unsigned char * )p_trans + sizeof( transaction_t );

    return p_trans;
}

/* Transaction reallocation */
transaction_t * ReallocTransaction( transaction_t *p_trans, ssize_t i_size )
{
    ssize_t i_allocsize = ( i_size + DATA_ALIGN - 1 ) & ~( DATA_ALIGN - 1 );
    ssize_t i_oldallocsize =
                 ( p_trans->i_size + DATA_ALIGN - 1 ) & ~( DATA_ALIGN - 1 );

    /* If a realloc isn't needed, return */
    if( i_allocsize == i_oldallocsize )
    {
        p_trans->i_size = i_size;
        return p_trans;
    }

    /* Return either the new p_trans or NULL */
    p_trans = realloc( p_trans, sizeof( transaction_t )
                                 + i_allocsize * sizeof( unsigned char ) );

    if( p_trans == NULL )
    {
        return NULL;
    }

    p_trans->i_size = i_size;
    p_trans->p_data = ( unsigned char * )p_trans + sizeof( transaction_t );

    return p_trans;
}

/* Append data to a transaction */
transaction_t * AppendToTransaction( transaction_t *p_trans, unsigned char *p_data, ssize_t i_size )
{
    transaction_t *p_newtrans;
    ssize_t i_oldsize = p_trans->i_size;

    p_newtrans = ReallocTransaction( p_trans, i_oldsize + i_size );

    if( p_newtrans == NULL )
    {
        return NULL;
    }

    memcpy( p_newtrans->p_data + i_oldsize, p_data, i_size );

    return p_newtrans;
}

/* Transaction deallocation */
void FreeTransaction( transaction_t *p_trans )
{
    free( p_trans );
}

